#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import json
from ... enums import *
from ... colors import Color
from ... properties import UNDEF
from ... drawing import Canvas, ClipState, GroupState


class JsonCanvas(Canvas):
    """Wrapper for buffered JSON drawing context."""
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of JsonCanvas."""
        
        # init buffers
        self._commands = []
        
        # init base
        super().__init__()
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_json_canvas_property_changed)
        
        # set overrides
        self.set_properties(overrides)
    
    
    def get_json(self):
        """
        Gets JSON string for current drawings.
        
        Returns:
            str
                Drawings JSON string.
        """
        
        return json.dumps({"commands": self._commands})
    
    
    def draw_arc(self, x, y, radius, start_angle, end_angle, clockwise=True):
        """
        Draws an arc of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the arc.
            
            start_angle: int or float
                Angle of the starting point in radians.
            
            end_angle: int or float
                Angle of the starting point in radians.
            
            clockwise: bool
                Direction of drawing between start and end point.
        """
        
        # store command
        self._store_command('draw_arc', {
            'x': x,
            'y': y,
            'radius': radius,
            'start_angle': start_angle,
            'end_angle': end_angle,
            'clockwise': clockwise})
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        # store command
        self._store_command('draw_circle', {
            'x': x,
            'y': y,
            'radius': radius})
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Draws an ellipse centered around given coordinates and fitting into the
        width and height.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # store command
        self._store_command('draw_ellipse', {
            'x': x,
            'y': y,
            'width': width,
            'height': height})
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # store command
        self._store_command('draw_line', {
            'x1': x1,
            'y1': y1,
            'x2': x2,
            'y2': y2})
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # disconnect points
        points = tuple((p[0], p[1]) for p in points)
        
        # store command
        self._store_command('draw_lines', {'points': points})
    
    
    def draw_path(self, path):
        """
        Draws given path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # get path dump
        path = json.loads(path.json())
        
        # store command
        self._store_command('draw_path', {'path': path})
    
    
    def draw_polygon(self, points):
        """
        Draws a closed polygon using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # disconnect points
        points = tuple((p[0], p[1]) for p in points)
        
        # store command
        self._store_command('draw_polygon', {'points': points})
    
    
    def draw_rect(self, x, y, width, height, radius=None):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corners starting from top-left.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # store command
        self._store_command('draw_rect', {
            'x': x,
            'y': y,
            'width': width,
            'height': height,
            'radius': radius})
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # store command
        self._store_command('draw_text', {
            'text': text,
            'x': x,
            'y': y,
            'angle': angle})
    
    
    def fill(self, color=None):
        """
        Fills current drawing region by specified or actual fill color.
        
        Args:
            color: pero.Color, (int,), str, None or UNDEF
                Specifies the fill color as an RGB or RGBA tuple, hex code, name
                or pero.Color. If not set, current fill color will be used.
        """
        
        # convert color
        color = Color.create(color).hex if color else None
        
        # store command
        self._store_command('fill', {'color': color})
    
    
    def view(self, x=None, y=None, width=None, height=None, relative=False):
        """
        Sets rectangular region currently used for drawing. This provides an
        easy way to draw complex graphics at specific position of the canvas
        without adjusting the coordinates of the graphics. It is achieved by
        changing the origin coordinates and the logical width and height of
        the canvas.
        
        Args:
            x: int or float
                X-coordinate of the top-left corner.
            
            y: int or float
                Y-coordinate of the top-left corner.
            
            width: int, float or None
                Full width of the viewport.
            
            height: int, float or None
                Full height of the viewport.
            
            relative: bool
                If set to True the new viewport is applied relative to current
                one.
        
        Returns:
            pero.ViewState
                Viewport state context manager.
        """
        
        # set to base
        state = super().view(x, y, width, height, relative)
        
        # store command
        self._store_command('view', {
            'x': x,
            'y': y,
            'width': width,
            'height': height,
            'relative': relative})
        
        return state
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # get path dump
        path = json.loads(path.json())
        
        # store command
        self._store_command('clip', {'path': path})
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """Removes last clipping path while keeping previous if any."""
        
        # store command
        self._store_command('unclip')
    
    
    def group(self, id_tag=None, class_tag=None):
        """
        Opens new drawing group.
        
        Args:
            id_tag: str
                Unique id of the group.
            
            class_tag:
                Class of the group.
        
        Returns:
            pero.GroupState
                Grouping state context manager.
        """
        
        # store command
        self._store_command('group', {
            'id_tag': id_tag,
            'class_tag': class_tag})
        
        # return state
        return GroupState(self)
    
    
    def ungroup(self):
        """Closes the last drawing group."""
        
        # store command
        self._store_command('ungroup')
    
    
    def _store_command(self, command, args=None):
        """Stores command and its parameters."""
        
        if args is None:
            args = {}
        
        self._commands.append((command, args))
    
    
    def _on_json_canvas_property_changed(self, evt):
        """Called after any property has changed."""
        
        # get value
        value = evt.new_value
        
        # convert UNDEF
        if value is UNDEF:
            value = str(UNDEF)
        
        # convert color
        elif isinstance(value, Color):
            value = value.hex
        
        # store command
        self._store_command('set_property', {
            'name': evt.name,
            'value': value,
            'raise_error': False})



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import weakref


class Proxy(object):
    """
    This class encapsulates given callback function to enable weak references
    to instance methods as well as functions.
    """
    
    
    def __init__(self, callback):
        """
        Initializes a new instance of Proxy.
        
        Args:
            callback: callable
                Callback function or method to be encapsulated.
        """
        
        # instance methods
        if hasattr(callback, '__self__'):
            self.obj = weakref.ref(callback.__self__)
            self.func = weakref.ref(callback.__func__)
        
        # direct function
        else:
            self.obj = None
            self.func = weakref.ref(callback)
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "Proxy(%s)" % self.callback
    
    
    def __call__(self, *args, **kwargs):
        """Calls defined callback."""
        
        self.callback(*args, **kwargs)
    
    
    def __eq__(self, other):
        """Compares two proxies for equality."""
        
        if not isinstance(other, Proxy):
            other = Proxy(other)
        
        if self.obj is other.obj:
            return self.func is other.func
        
        if self.obj is None or other.obj is None:
            return False
        
        return self.obj() is other.obj() and self.func() is other.func()
    
    
    def __ne__(self, other):
        """Compares two proxies for non-equality."""
        
        return not self.__eq__(other)
    
    
    @property
    def callback(self):
        """Gets the callback."""
        
        # direct function
        if self.obj is None:
            return self.func()
        
        # instance method
        obj = self.obj()
        if obj is not None:
            return getattr(obj, self.func().__name__)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import FrameProperty
from . glyph import Glyph


class Tooltip(Glyph):
    """
    Provides a base class for tooltip drawing.
    
    The pero.Tooltip classes can be used directly to draw tooltip or as
    descriptor to create a pero.Tooltip instances from real data and using the
    'clone' method and a data source.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the anchor.
        
        y: int, float or callable
            Specifies the y-coordinate of the anchor.
        
        x_offset: int, float or callable
            Specifies the x-axis shift from the anchor.
        
        y_offset: int, float or callable
            Specifies the y-axis shift from the anchor.
        
        anchor: pero.POSITION_COMPASS or callable
            Specifies the anchor position as any item from the
            pero.POSITION_COMPASS enum.
        
        clip: pero.Frame, callable, None or UNDEF
            Specifies the available drawing frame.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    
    x_offset = NumProperty(10)
    y_offset = NumProperty(10)
    
    anchor = EnumProperty(UNDEF, enum=POSITION_COMPASS)
    clip = FrameProperty(UNDEF)


class TextTooltip(Tooltip):
    """
    Provides a basic glyph to draw simple text box tooltip.
    
    Properties:
        
        text: str, callable, None or UNDEF
            Specifies the text to be drawn.
        
        text properties:
            Includes pero.TextProperties to specify the text properties. Some
            of them (e.g. alignment, angle, baseline) are set automatically.
        
        radius: int, float, tuple, callable or UNDEF
            Specifies the corner radius of the box as a single value or values
            for individual corners starting from top-left.
        
        padding: int, float, tuple, callable or UNDEF
            Specifies the inner space of the box as a single value or values for
            individual sides starting from top.
        
        line properties:
            Includes pero.LineProperties to specify the box outline.
        
        fill properties:
            Includes pero.FillProperties to specify the box fill.
    """
    
    text = StringProperty(UNDEF)
    
    font = Include(TextProperties)
    
    radius = QuadProperty(3)
    padding = QuadProperty(5)
    
    line = Include(LineProperties, line_color="#ccce")
    fill = Include(FillProperties, fill_color="#eeee")
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw tooltip."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        anchor = self.get_property('anchor', source, overrides)
        x_offset = self.get_property('x_offset', source, overrides)
        y_offset = self.get_property('y_offset', source, overrides)
        radius = self.get_property('radius', source, overrides)
        padding = self.get_property('padding', source, overrides)
        text = self.get_property('text', source, overrides)
        align = self.get_property('text_align', source, overrides)
        base = self.get_property('text_base', source, overrides)
        clip = self.get_property('clip', source, overrides)
        
        # check text
        if not text:
            return
        
        # set text
        canvas.set_text_by(self, source=source, overrides=overrides)
        
        # get size
        padding = padding or (0, 0, 0, 0)
        text_width, text_height = canvas.get_text_size(text)
        bgr_width = text_width + padding[1] + padding[3]
        bgr_height = text_height + padding[0] + padding[2]
        
        # shift anchor
        if anchor == POS_N:
            x -= 0.5 * bgr_width
        elif anchor == POS_NE:
            x -= bgr_width
        elif anchor == POS_E:
            x -= bgr_width
            y -= 0.5 * bgr_height
        elif anchor == POS_SE:
            x -= bgr_width
            y -= bgr_height
        elif anchor == POS_S:
            x -= 0.5 * bgr_width
            y -= bgr_height
        elif anchor == POS_SW:
            y -= bgr_height
        elif anchor == POS_W:
            y -= 0.5 * bgr_height
        elif anchor == POS_C:
            x -= 0.5 * bgr_width
            y -= 0.5 * bgr_height
        
        # apply offset
        x += x_offset
        y += y_offset
        
        # check clipping frame
        if clip:
            
            # shift horizontally if outside
            if x < clip.x1:
                x += 2*x_offset + bgr_width
            elif x + bgr_width > clip.x2 and x > clip.cx:
                x -= 2*x_offset + bgr_width
            
            # shift vertically if outside
            if y < clip.y1:
                y += 2*y_offset + bgr_height
            elif y + bgr_height > clip.y2 and y > clip.cy:
                y -= 2*y_offset + bgr_height
        
        # get text coords
        text_x = x + padding[3]
        text_y = y + padding[0]
        
        if align == TEXT_ALIGN_CENTER:
            text_x += 0.5*text_width
        elif align == TEXT_ALIGN_RIGHT:
            text_x += text_width
        
        if base == TEXT_BASE_MIDDLE:
            text_y += 0.5*text_height
        elif base == TEXT_BASE_BOTTOM:
            text_y += text_height
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # start drawing group
        canvas.group(tag, "tooltip")
        
        # draw background
        canvas.draw_rect(x, y, bgr_width, bgr_height, radius)
        
        # draw text
        canvas.draw_text(text, text_x, text_y)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# run all available tests
if __name__ == "__main__":
    
    import os.path
    import unittest
    
    suite = unittest.TestLoader().discover(os.path.dirname(__file__), pattern='test_*.py')
    unittest.TextTestRunner(verbosity=2).run(suite)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import sys
from . values import *
from . enum import Enum

# define line splitting character
LINE_SPLITTER = "\n"

# define default fonts
FONT_FACE_SERIF = "Times New Roman"
FONT_FACE_SANS = "Arial"
FONT_FACE_MONO = "Courier New"

# define font families
FONT_FAMILY_SERIF = SERIF
FONT_FAMILY_SANS = SANS
FONT_FAMILY_MONO = MONO

FONT_FAMILY = Enum(
    SERIF = FONT_FAMILY_SERIF,
    SANS = FONT_FAMILY_SANS,
    MONO = FONT_FAMILY_MONO)

# define font families default names
FONT_FAMILY_NAMES = {
    FONT_FAMILY_SERIF: (FONT_FACE_SERIF, 'times', 'Times', 'Times New Roman', 'Georgia', 'Liberation Serif'),
    FONT_FAMILY_SANS: (FONT_FACE_SANS, 'arial', 'Arial', 'Helvetica', 'Verdana', 'Liberation Sans', 'Nimbus Sans'),
    FONT_FAMILY_MONO: (FONT_FACE_MONO, 'courier', 'Courier', 'Courier New', 'Ubuntu Mono', 'Andale Mono', 'Liberation Mono')}

# define font styles
FONT_STYLE_NORMAL = NORMAL
FONT_STYLE_ITALIC = ITALIC

FONT_STYLE = Enum(
    NORMAL = FONT_STYLE_NORMAL,
    ITALIC = FONT_STYLE_ITALIC)

# define font weight
FONT_WEIGHT_NORMAL = NORMAL
FONT_WEIGHT_LIGHT = LIGHT
FONT_WEIGHT_BOLD = BOLD
FONT_WEIGHT_BLACK = BLACK
FONT_WEIGHT_HEAVY = HEAVY
FONT_WEIGHT_SEMIBOLD = SEMIBOLD
FONT_WEIGHT_MEDIUM = MEDIUM
FONT_WEIGHT_ULTRALIGHT = ULTRALIGHT
FONT_WEIGHT_THIN = THIN

FONT_WEIGHT = Enum(
    NORMAL = FONT_WEIGHT_NORMAL,
    LIGHT = FONT_WEIGHT_LIGHT,
    BOLD = FONT_WEIGHT_BOLD,
    BLACK = FONT_WEIGHT_BLACK,
    HEAVY = FONT_WEIGHT_HEAVY,
    SEMIBOLD = FONT_WEIGHT_SEMIBOLD,
    MEDIUM = FONT_WEIGHT_MEDIUM,
    ULTRALIGHT = FONT_WEIGHT_ULTRALIGHT,
    THIN = FONT_WEIGHT_THIN)

# define font weight light values
FONT_WEIGHTS_LIGHT = {
    FONT_WEIGHT_LIGHT,
    FONT_WEIGHT_ULTRALIGHT,
    FONT_WEIGHT_THIN}

# define font weight bold values
FONT_WEIGHTS_BOLD = {
    FONT_WEIGHT_BOLD,
    FONT_WEIGHT_BLACK,
    FONT_WEIGHT_HEAVY,
    FONT_WEIGHT_SEMIBOLD}

# define text align
TEXT_ALIGN_LEFT = LEFT
TEXT_ALIGN_CENTER = CENTER
TEXT_ALIGN_RIGHT = RIGHT

TEXT_ALIGN = Enum(
    LEFT = TEXT_ALIGN_LEFT,
    CENTER = TEXT_ALIGN_CENTER,
    RIGHT = TEXT_ALIGN_RIGHT)

# define text baseline
TEXT_BASE_TOP = TOP
TEXT_BASE_MIDDLE = MIDDLE
TEXT_BASE_BOTTOM = BOTTOM

TEXT_BASE = Enum(
    TOP = TEXT_BASE_TOP,
    MIDDLE = TEXT_BASE_MIDDLE,
    BOTTOM = TEXT_BASE_BOTTOM)

# define polar text rotation
TEXT_ROT_NONE = NONE
TEXT_ROT_FOLLOW = FOLLOW
TEXT_ROT_NATURAL = NATURAL
TEXT_ROT_FACEOUT = FACEOUT
TEXT_ROT_FACEIN = FACEIN

TEXT_ROTATION = Enum(
    NONE = TEXT_ROT_NONE,
    FOLLOW = TEXT_ROT_FOLLOW,
    NATURAL = TEXT_ROT_NATURAL,
    FACEOUT = TEXT_ROT_FACEOUT,
    FACEIN = TEXT_ROT_FACEIN)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import py5
from . canvas import Py5Canvas


class ShowSketch(py5.Sketch):
    """Custom helper py5.Sketch to draw and show given graphics."""
    
    
    def __init__(self, graphics, title, width, height, **options):
        """
        Initializes a new instance of ShowSketch.
        
        Args:
            graphics: pero.Graphics
                Graphics to draw.
            
            title: str
                Sketch window title.
            
            width: int
                Sketch size.
            
            height: int
                Sketch height.
            
            options: key:value pairs
                Optional properties passed into canvas creation.
        """
        
        # init buffs
        self._graphics = graphics
        self._title = title
        self._width = width
        self._height = height
        self._options = options
        
        # init base
        super().__init__()
    
    
    def settings(self):
        """Applies main sketch settings."""
        
        self.size(self._width, self._height)
    
    
    def setup(self):
        """Initializes canvas and draws current graphics."""
        
        # set title
        if self._title:
            self.window_title(self._title)
        
        # init py5 graphics
        pg = self.create_graphics(self._width, self._height)
        
        # init canvas
        canvas = Py5Canvas(pg, self, **self._options)
        
        # draw graphics
        with pg.begin_draw():
            self._graphics.draw(canvas)
        
        # finalize image
        self.image(pg, 0, 0)


class ExportSketch(py5.Sketch):
    """Custom helper py5.Sketch to export given graphics."""
    
    
    def __init__(self, graphics, path, width, height, **options):
        """
        Initializes a new instance of ExportSketch.
        
        Args:
            graphics: pero.Graphics
                Graphics to draw.
            
            path: str
                Export path
            
            width: int
                Sketch size.
            
            height: int
                Sketch height.
            
            options: key:value pairs
                Optional properties passed into canvas creation.
        """
        
        # init buffs
        self._graphics = graphics
        self._path = path
        self._width = width
        self._height = height
        self._options = options
        
        # init base
        super().__init__()
    
    
    def settings(self):
        """Applies main sketch settings."""
        
        self.size(self._width, self._height, py5.HIDDEN)
        self.pixel_density(1)
    
    
    def setup(self):
        """Initializes canvas and exports current graphics."""
        
        # export PDF
        if self._path.lower().endswith('.pdf'):
            self._export_pdf()
        
        # export SVG
        elif self._path.lower().endswith('.svg'):
            self._export_svg()
        
        # export raster
        else:
            self._export_raster()
        
        # close window
        self.exit_sketch()
    
    
    def _export_pdf(self):
        """Initializes canvas and exports current graphics as PDF."""
        
        # init py5 graphics
        pg = self.create_graphics(self._width, self._height, py5.PDF, self._path)
        
        # init canvas
        canvas = Py5Canvas(pg, self, **self._options)
        
        # draw graphics
        with self.begin_record(pg):
            self._graphics.draw(canvas)
    
    
    def _export_svg(self):
        """Initializes canvas and exports current graphics as SVG."""
        
        # init py5 graphics
        pg = self.create_graphics(self._width, self._height, py5.SVG, self._path)
        
        # init canvas
        canvas = Py5Canvas(pg, self, **self._options)
        
        # draw graphics
        with self.begin_record(pg):
            self._graphics.draw(canvas)
    
    
    def _export_raster(self):
        """Initializes canvas and exports current graphics as raster."""
        
        # init py5 graphics
        pg = self.create_graphics(self._width, self._height)
        
        # init canvas
        canvas = Py5Canvas(pg, self, **self._options)
        
        # draw graphics
        with pg.begin_draw():
            self._graphics.draw(canvas)
        
        # finalize image
        self.image(pg, 0, 0)
        
        # save image
        pg.save(self._path)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from . event import Event


class PropertyChangedEvt(Event):
    """
    Defines an event which is fired if any property of a pero.PropertySet
    was changed.
    
    Attributes:
        
        name: str
            Name of the changed property.
        
        old_value: any
            Original value.
        
        new_value: any
            New value.
    """
    
    TYPE = EVT_PROPERTY_CHANGED
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of PropertyChangedEvt."""
        
        self.name = None
        self.old_value = None
        self.new_value = None
        
        super().__init__(**kwargs)
    
    
    @classmethod
    def from_evt(cls, evt):
        """
        Initializes a new instance of given class by copying all current data.
        
        Args:
            evt: pero.PropertyChangedEvt
                Source event from which to copy the data.
        
        Returns:
            cls instance
                New instance of requested class.
        """
        
        return cls(
            name = evt.name,
            old_value = evt.old_value,
            new_value = evt.new_value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from .. enums import *
from .. properties import *
from .. drawing import Matrix, Path
from . glyph import Glyph


class Text(Glyph):
    """
    Defines a text glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the anchor.
        
        y: int, float or callable
            Specifies the y-coordinate of the anchor.
        
        angle properties:
            Includes pero.AngleProperties to specify the text angle.
        
        text: str, callable, None or UNDEF
            Specifies the text to be drawn.
        
        text properties:
            Includes pero.TextProperties to specify the font.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    angle = Include(AngleProperties)
    
    text = StringProperty(UNDEF)
    font = Include(TextProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        text = self.get_property('text', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # check data
        if not text:
            return
        
        # set text
        canvas.set_text_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_text(text, x, y, angle)


class Textbox(Text):
    """
    Defines a text box glyph.
    
    Properties:
        
        anchor: pero.POSITION_COMPASS or callable
            Specifies the anchor position as any item from the
            pero.POSITION_COMPASS enum. If set to UNDEF, anchor is derived from
            text alignment and baseline.
        
        radius: int, float, (int,), (float,) callable or UNDEF
            Specifies the corner radius as a single value or values for
            individual corners starting from top-left.
        
        padding: int, float, (int,), (float,) callable or UNDEF
            Specifies the inner space as a single value or values for individual
            sides starting from top.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    anchor = EnumProperty(UNDEF, enum=POSITION_COMPASS)
    
    radius = QuadProperty(0)
    padding = QuadProperty(5)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        anchor = self.get_property('anchor', source, overrides)
        radius = self.get_property('radius', source, overrides)
        padding = self.get_property('padding', source, overrides)
        text = self.get_property('text', source, overrides)
        align = self.get_property('text_align', source, overrides)
        base = self.get_property('text_base', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # check data
        if not text:
            return
        
        # set text
        canvas.set_text_by(self, source=source, overrides=overrides)
        
        # get size
        padding = padding or (0, 0, 0, 0)
        text_width, text_height = canvas.get_text_size(text)
        bgr_width = text_width + padding[1] + padding[3]
        bgr_height = text_height + padding[0] + padding[2]
        
        # get anchor
        if anchor is UNDEF:
            anchors = set(POSITION_COMPASS)
            
            if align == TEXT_ALIGN_LEFT or align == UNDEF:
                anchors = anchors.intersection(POSITION_COMPASS_LEFT)
            elif align == TEXT_ALIGN_CENTER:
                anchors = anchors.intersection(POSITION_COMPASS_CENTER)
            elif align == TEXT_ALIGN_RIGHT:
                anchors = anchors.intersection(POSITION_COMPASS_RIGHT)
            
            if base == TEXT_BASE_TOP or base == UNDEF:
                anchors = anchors.intersection(POSITION_COMPASS_TOP)
            elif base == TEXT_BASE_MIDDLE:
                anchors = anchors.intersection(POSITION_COMPASS_MIDDLE)
            elif base == TEXT_BASE_BOTTOM:
                anchors = anchors.intersection(POSITION_COMPASS_BOTTOM)
            
            anchor = anchors.pop()
        
        # get background coords
        bgr_x = x
        bgr_y = y
        
        if anchor == POS_N:
            bgr_x -= 0.5 * bgr_width
        elif anchor == POS_NE:
            bgr_x -= bgr_width
        elif anchor == POS_E:
            bgr_x -= bgr_width
            bgr_y -= 0.5 * bgr_height
        elif anchor == POS_SE:
            bgr_x -= bgr_width
            bgr_y -= bgr_height
        elif anchor == POS_S:
            bgr_x -= 0.5 * bgr_width
            bgr_y -= bgr_height
        elif anchor == POS_SW:
            bgr_y -= bgr_height
        elif anchor == POS_W:
            bgr_y -= 0.5 * bgr_height
        elif anchor == POS_C:
            bgr_x -= 0.5 * bgr_width
            bgr_y -= 0.5 * bgr_height
        
        # init text coords
        text_x = bgr_x + padding[3]
        text_y = bgr_y + padding[0]
        
        if align == TEXT_ALIGN_CENTER:
            text_x += 0.5*text_width
        elif align == TEXT_ALIGN_RIGHT:
            text_x += text_width
        
        if base == TEXT_BASE_MIDDLE:
            text_y += 0.5*text_height
        elif base == TEXT_BASE_BOTTOM:
            text_y += text_height
        
        if angle:
            dx = text_x - x
            dy = text_y - y
            text_x = x + dx * math.cos(angle) - dy * math.sin(angle)
            text_y = y + dx * math.sin(angle) + dy * math.cos(angle)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # start drawing group
        canvas.group()
        
        # draw angled rect
        if angle:
            path = Path()
            path.rect(bgr_x, bgr_y, bgr_width, bgr_height, radius)
            path.transform(Matrix().rotate(angle, x, y))
            canvas.draw_path(path)
        
        # strait rectangle
        else:
            canvas.draw_rect(bgr_x, bgr_y, bgr_width, bgr_height, radius)
        
        # draw text
        canvas.draw_text(text, text_x, text_y, angle)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from . interpols import *
from . scale import Scale


class ContinuousScale(Scale):
    """
    This type of scale tool is used to convert continuous numerical input values
    into continuous numerical output. This is typically used to convert real
    data units defined by 'in_range' property into device output units such as
    pixels defined by 'out_range' e.g. to draw plot axes and data series.
    
    The conversion is done in two steps: normalizing given value within the
    input range and converting the normalized value within the output range. For
    such functionality the scale contains a set of two interpolators. One is
    used for normalization 'normalizer' and the other for final conversion
    'converter'.
    
    To make a specific scale it is mostly sufficient to create an instance of
    ContinuousScale and set the 'normalizer' and/or 'converter' to provide
    specific conversion. There are already some predefined scales for convenient
    use like pero.LinScale, pero.LogScale and pero.PowScale.
    
    For the input values outside specified range the clipping can be applied by
    setting the 'clip' property to True. Otherwise such values will be
    extrapolated.
    
    To convert more data efficiently, the scale can convert not only a single
    value but whole sequence such as list or numpy.array at once. For custom
    scale be sure this functionality is supported by both interpolators.
    
    Properties:
        
        in_range: (float, float)
            Specifies the start and end value of the input range.
        
        out_range: (float, float)
            Specifies the start and end value of the output range.
        
        normalizer: pero.Interpol
            Specifies the normalization interpolator.
        
        converter: pero.Interpol
            Specifies the conversion interpolator.
        
        clip: bool
            Specifies whether the values outside current input range should be
            clipped (True) or interpolated (False).
    """
    
    in_range = TupleProperty((), intypes=(int, float), dynamic=False)
    out_range = TupleProperty((), intypes=(int, float), dynamic=False)
    
    normalizer = Property(UNDEF, types=(Interpol,), dynamic=False)
    converter = Property(UNDEF, types=(Interpol,), dynamic=False)
    clip = BoolProperty(False, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of ContinuousScale."""
        
        # init interpolators
        if 'normalizer' not in overrides:
            overrides['normalizer'] = LinInterpol()
        
        if 'converter' not in overrides:
            overrides['converter'] = LinInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # init indexes
        self._in_min = 0
        self._in_max = 1
        self._out_min = 0
        self._out_max = 1
        
        # update indexes
        self._on_continuous_scale_property_changed()
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_continuous_scale_property_changed)
    
    
    def scale(self, value, *args, **kwargs):
        """
        Returns corresponding output value for given input value.
        
        Args:
            value: int, float, (int,), (float,)
                Input value(s) to be scaled as a single number or a sequence of
                numbers.
        
        Returns:
            int, float or numpy.ndarray
                Scaled value(s).
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return self._scale_array(value)
        
        # clip values outside
        if self.clip:
            
            if value <= self.in_range[self._in_min]:
                return self.out_range[self._in_min]
            
            if value >= self.in_range[self._in_max]:
                return self.out_range[self._in_max]
        
        # normalize value
        norm = self.normalizer.normalize(value, self.in_range[0], self.in_range[1])
        
        # convert normalized value
        return self.converter.denormalize(norm, self.out_range[0], self.out_range[1])
    
    
    def invert(self, value, *args, **kwargs):
        """
        Returns corresponding input value for given output value.
        
        Args:
            value: int, float, (int,), (float,)
                Output value(s) to be inverted as a single number or a sequence
                of numbers.
        
        Returns:
            int, float or numpy.ndarray
                Inverted value(s).
        """
        
        # apply array inverting
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return self._invert_array(value)
        
        # clip values outside
        if self.clip:
            
            if value <= self.out_range[self._out_min]:
                return self.in_range[self._out_min]
            
            if value >= self.out_range[self._out_max]:
                return self.in_range[self._out_max]
        
        # normalize value
        norm = self.converter.normalize(value, self.out_range[0], self.out_range[1])
        
        # convert normalized value
        return self.normalizer.denormalize(norm, self.in_range[0], self.in_range[1])
    
    
    def normalize(self, value, *args, **kwargs):
        """
        Returns normalized value (in range 0 to 1) for given input value.
        
        Args:
            value: int, float, (int,), (float,)
                Input value(s) to be normalized as a single number or a sequence
                of numbers.
        
        Returns:
            int, float or numpy.ndarray
                Normalized value(s).
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return self._normalize_array(value)
        
        # clip values outside
        if self.clip:
            
            if value <= self.in_range[self._in_min]:
                return float(self._in_min)
            
            if value >= self.in_range[self._in_max]:
                return float(self._in_max)
        
        # normalize value
        return self.normalizer.normalize(value, self.in_range[0], self.in_range[1])
    
    
    def _scale_array(self, value):
        """Returns corresponding output array for given input array."""
        
        # check array
        if not isinstance(value, numpy.ndarray):
            value = numpy.array(value)
        
        # clip values
        if self.clip:
            numpy.clip(value, self.in_range[self._in_min], self.in_range[self._in_max], out=value)
        
        # normalize values
        norm = self.normalizer.normalize(value, self.in_range[0], self.in_range[1])
        
        # convert normalized values
        denorm = self.converter.denormalize(norm, self.out_range[0], self.out_range[1])
        
        return denorm
    
    
    def _invert_array(self, value):
        """Returns corresponding input array for given output array."""
        
        # check array
        if not isinstance(value, numpy.ndarray):
            value = numpy.array(value)
        
        # clip values
        if self.clip:
            numpy.clip(value, self.out_range[self._out_min], self.out_range[self._out_max], out=value)
        
        # normalize values
        norm = self.converter.normalize(value, self.out_range[0], self.out_range[1])
        
        # convert normalized values
        denorm = self.normalizer.denormalize(norm, self.in_range[0], self.in_range[1])
        
        return denorm
    
    
    def _normalize_array(self, value):
        """Returns corresponding normalized array for given input array."""
        
        # check array
        if not isinstance(value, numpy.ndarray):
            value = numpy.array(value)
        
        # normalize values
        norm = self.normalizer.normalize(value, self.in_range[0], self.in_range[1])
        
        # clip values
        if self.clip:
            numpy.clip(norm, 0., 1., out=norm)
        
        return norm
    
    
    def _on_continuous_scale_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # check in_range
        if evt is None or evt.name in 'in_range':
            self._in_min = 0
            self._in_max = 1
            if self.in_range and self.in_range[0] > self.in_range[1]:
                self._in_min = 1
                self._in_max = 0
        
        # check out_range
        if evt is None or evt.name == 'out_range':
            self._out_min = 0
            self._out_max = 1
            if self.out_range and self.out_range[0] > self.out_range[1]:
                self._out_min = 1
                self._out_max = 0


class LinScale(ContinuousScale):
    """Continuous scale with linear transformation."""
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of LinScale."""
        
        # init interpolators
        overrides['normalizer'] = LinInterpol()
        overrides['converter'] = LinInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # lock interpolators
        self.lock_property('normalizer')
        self.lock_property('converter')


class LogScale(ContinuousScale):
    """Continuous scale with logarithmic transformation."""
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of LogScale."""
        
        # init interpolators
        overrides['normalizer'] = LogInterpol()
        overrides['converter'] = LinInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # lock interpolators
        self.lock_property('normalizer')
        self.lock_property('converter')


class PowScale(ContinuousScale):
    """
    Continuous scale with exponential transformation.
    
    Properties:
        
        power: int or float
            Specifies the exponent to be used.
    """
    
    power = FloatProperty(1, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of PowScale."""
        
        # init interpolators
        overrides['normalizer'] = PowInterpol()
        overrides['converter'] = LinInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # lock interpolators
        self.lock_property('normalizer')
        self.lock_property('converter')
        
        # set power
        self.normalizer.power = self.power
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_pow_scale_property_changed)
    
    
    def _on_pow_scale_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # check power
        if evt is None or evt.name == 'power':
            self.normalizer.power = self.power



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from .. formatters import Formatter


class Ticker(PropertySet):
    """
    Abstract base class for various types of ticks and labels generators, which
    are typically used to generate nice looking ticks, labels and grids for
    charts.
    
    All derived classes have to implement the 'make_ticks' method, which is
    then automatically used to calculate major and minor ticks.
    
    Labels are automatically created from major ticks using specified
    'formatter'. The formatter 'domain' is automatically updated by current
    range. Its 'precision', however, must be updated manually in derived classes
    as needed.
    
    For some cases it is useful to hide either the first or the last tick and
    label, such as for radial axes. This can be achieved by setting the
    'hide_first' or 'hide_last' properties.
    
    Properties:
        
        start: int or float
            Specifies the start value of the range in data units.
        
        end: int or float
            Specifies the end value of the range in data units.
        
        formatter: pero.Formatter
            Specifies the formatter to be used to format labels.
        
        hide_first: bool
            Specifies whether the first tick and label should be removed.
        
        hide_last: bool
            Specifies whether the last tick and label should be removed.
    """
    
    start = NumProperty(UNDEF, dynamic=False)
    end = NumProperty(UNDEF, dynamic=False)
    
    formatter = Property(UNDEF, types=(Formatter,), dynamic=False)
    hide_first = BoolProperty(False, dynamic=False)
    hide_last = BoolProperty(False, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Ticker."""
        
        super().__init__(**overrides)
        
        # init buffers
        self._major_ticks = ()
        self._minor_ticks = ()
        self._labels = ()
        self._is_dirty = True
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_ticker_property_changed)
    
    
    def initialize(self):
        """Recalculates internal values to current settings if necessary."""
        
        # skip if not necessary
        if not self._is_dirty:
            return
        
        # reset buffers
        self._major_ticks = ()
        self._minor_ticks = ()
        self._labels = ()
        
        # update formatter
        self.formatter.domain = max(abs(self.start), abs(self.end))
        
        # make ticks
        major_ticks, minor_ticks = self.make_ticks()
        self._major_ticks = tuple(major_ticks)
        self._minor_ticks = tuple(minor_ticks)
        
        # reset flag
        self._is_dirty = False
    
    
    def make_ticks(self):
        """
        Generates ticks according to current settings. This method must be
        overridden in derived classes to provide specific mechanism to calculate
        major and minor ticks.
        
        Returns:
            (float,), (float,)
                Generated major and minor ticks.
        """
        
        raise NotImplementedError("The 'make_ticks' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def major_ticks(self):
        """
        Gets calculated major ticks.
        
        Returns:
            (float,)
                Major ticks values.
        """
        
        # recalculate values
        if self._is_dirty:
            self.initialize()
        
        # get ticks
        ticks = self._major_ticks
        
        if self.hide_first and len(ticks) > 0:
            ticks = ticks[1:]
        
        if self.hide_last and len(ticks) > 0:
            ticks = ticks[:-1]
        
        return ticks
    
    
    def minor_ticks(self):
        """
        Gets calculated minor ticks.
        
        Returns:
            (float,)
                Minor ticks values in data units.
        """
        
        # recalculate values
        if self._is_dirty:
            self.initialize()
        
        # get ticks
        ticks = self._minor_ticks
        
        if self.hide_first and len(ticks) > 0:
            ticks = ticks[1:]
        
        if self.hide_last and len(ticks) > 0:
            ticks = ticks[:-1]
        
        return ticks
    
    
    def labels(self):
        """
        Gets formatted labels for major ticks.
        
        Returns:
            (str,)
                Formatted major ticks labels.
        """
        
        # recalculate values
        if self._is_dirty:
            self.initialize()
        
        # format labels
        if not self._labels:
            self._labels = tuple(map(self.format, self._major_ticks))
        
        # get labels
        labels = self._labels
        
        if self.hide_first and len(labels) > 0:
            labels = labels[1:]
        
        if self.hide_last and len(labels) > 0:
            labels = labels[:-1]
        
        return labels
    
    
    def suffix(self):
        """
        Gets current suffix to be added to axis title (e.g. 10^5).
        
        Returns:
            str
                Labels suffix.
        """
        
        # recalculate values
        if self._is_dirty:
            self.initialize()
        
        # get current suffix
        return self.formatter.suffix()
    
    
    def format(self, value):
        """
        Formats given label value using current formatter.
        
        Args:
            value: any
                A value in data units to be formatted for label.
        
        Return:
            str
                Formatted label value.
        """
        
        # recalculate values
        if self._is_dirty:
            self.initialize()
        
        # apply formatting
        return self.formatter.format(value)
    
    
    def beautify(self, start, end):
        """
        Calculates visually nice range to cover given range.
        
        This method should be overridden in derived classes to provide specific
        mechanism to beautify given range.
        
        Args:
            start: int or float
                Minimum value of the range.
            
            end: int or float
                Maximum value of the range.
        
        Returns:
            (float, float)
                New extended range.
        """
        
        return start, end
    
    
    def _on_ticker_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # mark as dirty
        self._is_dirty = True
        
        # reset buffers
        self._major_ticks = ()
        self._minor_ticks = ()
        self._labels = ()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# prepare modules
from . import np

# import main objects
from . fonts import Font, FontManager, FONTS
from . frame import Frame, FrameProperty
from . graphics import Graphics
from . canvas import Canvas, ClipState, GroupState, ViewState
from . matrix import Matrix
from . bezier import Bezier
from . path import Path
from . arch import Arch
from . layout import Layout, Row, Column, Cell

# import utils
from . utils import *

# register new properties
from .. import properties
properties.FrameProperty = FrameProperty



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. events import PropertyChangedEvt
from . undefined import UNDEF


class Property(object):
    """
    Defines a generic property used in all pero.PropertySet classes. The main
    purpose is to provide reasonable default value and possibility to allow
    certain types or values only. Specific derived implementations can be
    created to provide custom value checking and parsing by overwriting the
    'parse' and 'clone' methods.
    
    If allowed by the 'dynamic' attribute the actual value can be a function or
    method, which is expected to give the final value from provided data source.
    This allows properties to be dynamic and provide specific values based on
    actual data.
    
    Attributes:
        
        name: str, (read-only)
            Property name.
        
        default: any
            Default value.
        
        types: (Type,) (read-only)
            Specifies the allowed types for values. If empty, specific type is
            not required for values.
        
        dynamic: bool (read-only)
            Specifies whether the property can be defined as dynamic i.e. the
            actual value of the property is a function, which is expected to
            give the final value from provided data source. 
        
        nullable: bool (read-only)
            Specifies whether the property value can be set to None.
    """
    
    
    def __init__(self, default=UNDEF, types=(), dynamic=True, nullable=False, name=None):
        """
        Initializes a new instance of Property.
        
        Args:
            default: any
                Default value.
            
            types: (Type,)
                Specifies allowed types for values. If empty, specific type is
                not required for values.
            
            dynamic: bool
                Specifies whether the property can be defined as dynamic i.e.
                the actual value of the property is a function, which is
                expected to give the final value from provided data source.
            
            nullable: bool
                Specifies whether the property value can be set to None.
            
            name: str
                Property name.
        """
        
        self._name = name
        self._types = types
        self._dynamic = bool(dynamic)
        self._nullable = bool(nullable)
        
        # check types
        if isinstance(types, list):
            self._types = tuple(types)
        
        elif not isinstance(types, tuple):
            self._types = (types,)
        
        # check default
        self.parse(default)
        
        # set default
        self._default = default
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return "%s(%s:%s)" % (self.__class__.__name__, self._name, self._default)
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return self.__str__()
    
    
    def __get__(self, obj, objtype=None):
        """
        Depending on the context, this method returns either the property value
        or the property itself.
        
        If the 'obj' is set to an instance of pero.PropertySet, actual
        property value is returned. If such property is not yet set in the obj
        instance, the default value is returned and set to the instance as
        well.
        
        If the 'obj' is None but 'objtype' is set, the property itself is
        returned.
        """
        
        # get instance value
        if obj is not None:
            
            # check name
            name = self._name
            if not name:
                raise RuntimeError('The property name is not set!')
            
            # get known value
            if name in obj.__dict__:
                return obj.__dict__[name]
            
            # initialize by default value
            else:
                new_value = self.parse(self._default)
                obj.__dict__[name] = new_value
                return new_value
        
        # get property itself
        if objtype is not None:
            return self
        
        message = "Cannot get the '%s' property!" % self._name
        raise ValueError(message)
    
    
    def __set__(self, obj, value):
        """
        Depending on the context, this method parses and sets given value to
        the instance or changes the default value of the property.
        
        If the 'obj' is set to an instance of PropertySet, given value is set to
        the instance. If the value is different from the one previously set,
        the pero.EVT_PROPERTY_CHANGED event is fired with current property
        name and the old and the new values.
        
        If the 'obj' is None given value is set as default value of the property
        itself.
        """
        
        # parse value
        new_value = self.parse(value)
        
        # set instance value
        if obj is not None:
            
            # check name
            name = self._name
            if not name:
                raise RuntimeError('The property name is not set!')
            
            # check lock
            if obj.is_property_locked(name):
                message = "The property '%s' is locked and cannot be changed!" % name
                raise AttributeError(message)
            
            # check undefinable
            if value is UNDEF and obj.is_property_held(name):
                return
            
            # get old value
            if name in obj.__dict__:
                old_value = obj.__dict__[name]
            else:
                old_value = self.parse(self._default)
                obj.__dict__[name] = old_value
            
            # compare values
            if type(old_value) != type(new_value):
                replace = True
            elif isinstance(old_value, numpy.ndarray):
                replace = not numpy.array_equal(old_value, new_value)
            else:
                replace = old_value != new_value
            
            # set new value and raise changed event
            if replace:
                obj.__dict__[name] = new_value
                obj.fire(PropertyChangedEvt(name=name, old_value=old_value, new_value=new_value))
        
        # set class default
        else:
            self._default = value
    
    
    @property
    def name(self):
        """
        Gets property name.
        
        Returns:
            str
                Property name.
        """
        
        return self._name or self.__class__.name
    
    
    @property
    def types(self):
        """
        Gets allowed types.
        
        Returns:
            (Type,)
                Allowed types.
        """
        
        return self._types
    
    
    @property
    def dynamic(self):
        """
        Gets the value indicating whether the property can be dynamic.
        
        Returns:
            bool
                True if the property can be dynamic, False otherwise.
        """
        
        return self._dynamic
    
    
    @property
    def nullable(self):
        """
        Gets the value indicating whether the property value can be set to None.
        
        Returns:
            bool
                True if the property can be set to None, False otherwise.
        """
        
        return self._nullable
    
    
    @property
    def default(self):
        """
        Gets default value.
        
        Returns:
            any
                Default value.
        """
        
        return self._default
    
    
    def parse(self, value):
        """
        Validates and converts given value.
        
        Args:
            value: any
                Value to be parsed.
        
        Returns:
            any
                Parsed value.
        """
        
        # check allowed types
        if self._types and isinstance(value, self._types):
            return value
        
        # check nullable
        if not self._nullable and value is None:
            message = "Value of the '%s' property cannot be None!" % self.name
            raise ValueError(message)
        
        # check dynamic
        if not self._dynamic and callable(value):
            message = "Value of the '%s' property cannot be dynamic!" % self.name
            raise ValueError(message)
        
        # allow None, UNDEF, dynamic or untyped
        if value is None or value is UNDEF or callable(value) or not self._types:
            return value
        
        # try to convert into single defined type
        if len(self._types) == 1:
            try: return self._types[0](value)
            except: pass
        
        # cannot convert value
        types = "|".join(x.__name__ for x in self._types)
        message = "Value of the '%s' property must be of type (%s)! -> %s" % (self.name, types, type(value))
        raise TypeError(message)
    
    
    def clone(self, **kwargs):
        """
        Creates a clone of current property. Note that any custom property,
        which defines its own attributes should overwrite this method and
        provide its own attributes as keyword arguments to this base method.
        
        Args:
            kwargs: {key:value,}
                Arguments to overwrite in the clone.
        
        Returns:
            pero.Property
                Cloned property.
        """
        
        if 'name' not in kwargs:
            kwargs['name'] = self.name
        
        if 'default' not in kwargs:
            kwargs['default'] = self.default
        
        if 'types' not in kwargs:
            kwargs['types'] = self.types
        
        if 'dynamic' not in kwargs:
            kwargs['dynamic'] = self.dynamic
        
        if 'nullable' not in kwargs:
            kwargs['nullable'] = self.nullable
        
        return self.__class__(**kwargs)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . control import Control


class View(object):
    """Abstract base class for specific backend implementations of views."""
    
    
    def __init__(self):
        """Initializes a new instance of View."""
        
        self._control = None
    
    
    @property
    def control(self):
        """
        Gets current control.
        
        Returns:
            pero.Control or None
        """
        
        return self._control
    
    
    def set_control(self, control):
        """
        Sets current control.
        
        Args:
            control: pero.Control or None
                Specific control to be set.
        """
        
        # reset control
        if control is None:
            
            # remove parent link
            if self._control is not None:
                self._control.set_parent(None)
            
            # reset control
            self._control = None
            return
        
        # check control
        if not isinstance(control, Control):
            message = "Control must be of type 'pero.Control'! -> %s" % type(control)
            raise TypeError(message)
        
        # set control
        self._control = control
        
        # set parent
        self._control.set_parent(self)
    
    
    def set_cursor(self, cursor):
        """
        This method should be overridden to provide specific mechanism to set
        given mouse cursor.
        
        Args:
            cursor: pero.CURSOR
                Cursor type to be set. The value must be an item from the
                pero.CURSOR enum.
        """
        
        pass
    
    
    def set_tooltip(self, text):
        """
        This method should be overridden to provide specific mechanism to set
        given text as system tooltip.
        
        Args:
            text: str
                Tooltip text to be shown.
        """
        
        pass
    
    
    def refresh(self):
        """Redraws current control graphics."""
        
        self.draw_control()
    
    
    def draw_control(self):
        """
        This method should be overridden to provide specific drawing mechanism
        and canvas creation to draw current control graphics.
        """
        
        raise NotImplementedError("The 'draw_control' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def draw_overlay(self, func=None, view=None, **kwargs):
        """
        This method should be overridden to provide specific drawing mechanism
        and canvas creation to finally call given function to draw cursor rubber
        band overlay over the current graphics.
        
        Specified function is expected to be called with a canvas as the first
        argument followed by given overrides (i.e. func(canvas, **kwargs)).
        
        Calling this method without any parameter should clear current overlay.
        
        Args:
            func: callable or None
                Drawing function to be called to draw the overlay. If set to
                None, current overlay will be cleared.
            
            view: (int, int, int, int) or None
                Rectangle defined as (x, y, width, height) used to shift origin
                of drawing canvas submitted to given drawing function.
                
            kwargs: str:any pairs
                Keyword arguments, which should be provided to the given drawing
                function.
        """
        
        raise NotImplementedError("The 'draw_overlay' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def clear_overlay(self):
        """Clears current overlay."""
        
        self.draw_overlay()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from . import utils
from . frame import Frame


class Arch(object):
    """Represents a simple arch."""
    
    
    def __init__(self, x, y, radius, start_angle, end_angle, clockwise):
        """
        Initializes new instance of arch by its center coordinates, radius,
        start and end angle.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the arch.
            
            start_angle: float
                Start angle in radians.
            
            end_angle: float
                End angle in radians.
            
            clockwise: bool
                Specifies the direction of drawing. If set to True, the arch
                will be drawn in the clockwise direction.
        """
        
        self._x = float(x)
        self._y = float(y)
        self._radius = float(radius)
        self._start_angle = utils.normal_angle(start_angle)
        self._end_angle = utils.normal_angle(end_angle)
        self._clockwise = bool(clockwise)
        
        self._bbox = None
        self._angle = None
        self._start_point = None
        self._end_point = None
        self._mid_point = None
    
    
    @property
    def center(self):
        """
        Gets the center of the arch.
        
        Returns:
            (float, float)
                Arch center coordinates.
        """
        
        return self._x, self._y
    
    
    @property
    def start(self):
        """
        Gets the starting point of the arch.
        
        Returns:
            (float, float)
                Arch start coordinates.
        """
        
        if self._start_point is None:
            self._start_point = self.angle_as_point(self._start_angle)
        
        return self._start_point
    
    
    @property
    def end(self):
        """
        Gets the end point of the arch.
        
        Returns:
            (float, float)
                Arch end coordinates.
        """
        
        if self._end_point is None:
            self._end_point = self.angle_as_point(self._end_angle)
        
        return self._end_point
    
    
    @property
    def middle(self):
        """
        Gets the middle angle point of the arch.
        
        Returns:
            (float, float)
                Arch middle angle coordinates.
        """
        
        if self._mid_point is None:
            angle = abs(self.angle() / 2)
            angle += self._start_angle if self._clockwise else self._end_angle
            self._mid_point = self.angle_as_point(angle)
        
        return self._mid_point
    
    
    @property
    def radius(self):
        """
        Gets the arch radius.
        
        Returns:
            float
                Radius.
        """
        
        return self._radius
    
    
    @property
    def start_angle(self):
        """
        Gets the start angle in radians.
        
        Returns:
            float
                Start angle in radians.
        """
        
        return self._start_angle
    
    
    @property
    def end_angle(self):
        """
        Gets the end angle in radians.
        
        Returns:
            float
                Start end in radians.
        """
        
        return self._end_angle
    
    
    @property
    def clockwise(self):
        """
        Gets the arch direction indicator.
        
        Returns:
            float
                True if clockwise, False otherwise.
        """
        
        return self._clockwise
    
    
    def bbox(self):
        """
        Calculates arch bounding box.
        
        Returns:
            pero.Frame or None
                Arch bounding box.
        """
        
        if self._bbox is None:
            
            # init frame
            self._bbox = Frame(*self.start)
            self._bbox.extend(*self.end)
            
            # add extremes if existing
            if self.contains_angle(0):
                self._bbox.extend(self._x + self._radius, self._y)
            
            if self.contains_angle(0.5*numpy.pi):
                self._bbox.extend(self._x, self._y + self._radius)
            
            if self.contains_angle(numpy.pi):
                self._bbox.extend(self._x - self._radius, self._y)
            
            if self.contains_angle(1.5*numpy.pi):
                self._bbox.extend(self._x, self._y - self._radius)
        
        return self._bbox
    
    
    def length(self):
        """Returns actual length of the arch."""
        
        return self._radius * self.angle()
    
    
    def angle(self):
        """
        Gets the angle between start and end.
        
        Returns:
            float
                Arch angle in radians.
        """
        
        if self._angle is None:
            self._angle = utils.angle_difference(self._start_angle, self._end_angle, self._clockwise)
        
        return self._angle
    
    
    def segment_area(self):
        """
        Calculates area of the arch segment, which is the area of the arch
        without the triangle, which goes to the center.
        
        Returns:
            float
                Sector area.
        """
        
        theta = abs(self.angle())
        return self._radius * self._radius / 2 * (theta - numpy.sin(theta))
    
    
    def sector_area(self):
        """
        Calculates area of the arch sector, which is the area of the arch plus
        the triangle, which goes to the center.
        
        Returns:
            float
                Sector area.
        """
        
        return self._radius * self._radius / 2 * abs(self.angle())
    
    
    def angle_as_point(self, angle):
        """
        Calculates coordinates of the point with given angle using current arch
        center and radius.
        
        Args:
            angle: float
                Angle in radians.
        
        Returns:
            (float, float)
                Coordinates of the point.
        """
        
        x = self._x + self._radius * numpy.cos(angle)
        y = self._y + self._radius * numpy.sin(angle)
        
        return x, y
    
    
    def angle_from_start(self, x, y):
        """
        Calculates angle between the arch start and given point respecting
        current direction.
        
        Args:
            x: float
                X-coordinate of the point.
            
            y: float
                Y-coordinate of the point.
        
        Returns:
            float
                Angle in radians.
        """
        
        angle = self.point_as_angle(x, y)
        diff = utils.angle_difference(self._start_angle, angle, self._clockwise)
        
        return diff
    
    
    def point_as_angle(self, x, y):
        """
        Calculates angle of given point.
        
        Args:
            x: float
                X-coordinate of the point.
            
            y: float
                Y-coordinate of the point.
        
        Returns:
            float
                Angle in radians.
        """
        
        angle = utils.inclination((self._x, self._y), (x, y))
        return utils.normal_angle(angle)
    
    
    def contains_point(self, x, y, inside=False):
        """
        Returns True if given point lays on the arch.
        
        Args:
            x: float
                X-coordinate of the point.
            
            y: float
                Y-coordinate of the point.
            
            inside: bool
                If set to True, point must be inside the arch and cannot be
                equal to start or end points.
        
        Returns:
            bool
                Returns True if the arch contains the point.
        """
        
        angle = self.point_as_angle(x, y)
        return self.contains_angle(angle, inside)
    
    
    def contains_angle(self, angle, inside=False):
        """
        Returns True if given angle is between start and end angle of the arch.
        
        Args:
            angle: float
                The angle to check.
            
            inside: bool
                If set to True, angle must be inside the arch and cannot be
                equal to start or end angle.
        
        Returns:
            bool
                Returns True if the arch contains the angle.
        """
        
        angle = utils.normal_angle(angle)
        diff = utils.angle_difference(self._start_angle, angle, self._clockwise)
        
        if inside:
            return abs(diff) < abs(self.angle())
        
        return abs(diff) <= abs(self.angle())
    
    
    def intersect_circle(self, x, y, radius, inside=False):
        """
        Calculates all intersection points between current arch and given
        circle. The points are sorted by increasing angle from the arch start.
        
        Args:
            x: float
                X-coordinate of the circle center.
            
            y: float
                Y-coordinate of the circle center.
            
            radius: float
                Radius of the circle.
            
            inside: bool
                If set to True, intersection points must be inside the arch
                and cannot be equal to edge points.
        
        Returns:
            ((float, float),)
                Coordinates of all intersection points.
        """
        
        # calc circles intersection
        points = utils.intersect_circles((self._x, self._y), self._radius, (x, y), radius)
        if not points:
            return ()
        
        # get points inside current arch
        points = [p for p in points if self.contains_point(p[0], p[1], inside)]
        
        # sort by angle from start
        points.sort(key=lambda p: abs(self.angle_from_start(p[0], p[1])))
        
        return points
    
    
    def intersect_arch(self, arch, inside=False):
        """
        Calculates all intersection points between current arch and given arch.
        
        Args:
            arch: pero.Arch
                Arch to intersect.
            
            inside: bool
                If set to True, intersection points must be inside the arch
                and cannot be equal to edge points.
        
        Returns:
            ((float, float),)
                Coordinates of all intersection points.
        """
        
        # calc circle intersection
        points = self.intersect_circle(arch.center[0], arch.center[1], arch.radius, inside)
        
        # get points inside given arch
        points = [p for p in points if arch.contains_point(p[0], p[1], inside)]
        
        return points
    
    
    @staticmethod
    def from_points(center, start_point, end_point, radius, clockwise):
        """
        Initializes new instance of arch by its center coordinates, start and
        end point coordinates, radius and direction.
        
        Args:
            center: (float, float)
                Coordinates of the center.
            
            start_point: (float, float)
                Coordinates of the starting point.
            
            end_point: (float, float)
                Coordinates of the end point.
            
            radius: int or float
                Radius of the arch.
            
            clockwise: bool
                Specifies the direction of drawing. If set to True, the arch
                will be drawn in the clockwise direction.
        """
        
        start_angle = utils.inclination(center, start_point)
        end_angle = utils.inclination(center, end_point)
        
        return Arch(center[0], center[1], radius, start_angle, end_angle, clockwise)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
import html
from ... properties import *
from ... drawing import Canvas, Path, Matrix, ClipState, GroupState
from . enums import *

# define constants
_INDENT = "  "


class SVGCanvas(Canvas):
    """Wrapper for SVG drawing context."""
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of SVGCanvas."""
        
        # init buffers
        self._commands = []
        self._filters = {}
        self._clips = {}
        self._indent = _INDENT
        
        self._pen_attrs = {
            'stroke': None,
            'stroke-opacity': None,
            'stroke-width': None,
            'stroke-dasharray': None,
            'stroke-linecap': None,
            'stroke-linejoin': None}
        
        self._brush_attrs = {
            'fill': None,
            'fill-opacity': None}
        
        self._font_attrs = {
            'font-family': None,
            'font-size': None,
            'font-style': None,
            'font-weight': None,
            'fill': None,
            'fill-opacity': None,
            'filter': None}
        
        self._font_descent = 0
        
        # init base
        super().__init__(**overrides)
        
        # init canvas
        self._update_pen()
        self._update_brush()
        self._update_text()
        
        # bind events
        self.bind(EVT_PEN_CHANGED, self._update_pen)
        self.bind(EVT_BRUSH_CHANGED, self._update_brush)
        self.bind(EVT_TEXT_CHANGED, self._update_text)
    
    
    def get_xml(self):
        """
        Gets full XML for current drawings. The output already contains the main
        XML tags so it can be directly saved into a file.
        
        Returns:
            str
                Drawings XML.
        """
        
        # make xml
        xml = '<?xml version="1.0"?>'
        xml += '\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.0//EN" "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">\n\n'
        xml += self.get_svg()
        
        return xml
    
    
    def get_svg(self):
        """
        Gets SVG markup for current drawings. The output does not contain the
        main XML tags so it can be directly used within HTML documents.
        
        Returns:
            str
                Drawings SVG markup.
        """
        
        # init xml
        xml = '<svg xmlns="http://www.w3.org/2000/svg" width="%s" height="%s">\n' % (self.width, self.height)
        
        # add defs
        if self._filters or self._clips:
            xml += "  <defs>\n"
            
            # add filters
            if self._filters:
                xml += "    %s\n" % "\n    ".join(self._filters.values())
            
            # add clip paths
            if self._clips:
                xml += "    %s\n" % "\n    ".join(self._clips.values())
            
            xml += "  </defs>\n"
        
        # add drawings
        xml += "\n".join(self._commands)
        
        # finish xml
        xml += '\n</svg>'
        
        return xml
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        radius = self._scale * radius
        
        # get pen and brush
        pen = self._get_pen_attrs()
        brush = self._get_brush_attrs()
        
        # make command
        command = self._indent + '<circle cx="%s" cy="%s" r="%s" %s %s />' % (x, y, radius, pen, brush)
        
        # add command
        self._commands.append(command)
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Draws an ellipse centered around given coordinates and fitting into the
        width and height.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # get pen and brush
        pen = self._get_pen_attrs()
        brush = self._get_brush_attrs()
        
        # make command
        command = self._indent + '<ellipse cx="%s" cy="%s" rx="%s" ry="%s" %s %s />' % (x, y, 0.5*width, 0.5*height, pen, brush)
        
        # add command
        self._commands.append(command)
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # apply scaling and offset
        x1 = self._scale * (x1 + self._offset[0])
        y1 = self._scale * (y1 + self._offset[1])
        x2 = self._scale * (x2 + self._offset[0])
        y2 = self._scale * (y2 + self._offset[1])
        
        # get pen
        pen = self._get_pen_attrs()
        
        # make command
        command = self._indent + '<line x1="%s" y1="%s" x2="%s" y2="%s" %s />' % (x1, y1, x2, y2, pen)
        
        # add command
        self._commands.append(command)
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # format
        points = ("%s,%s" % (x,y) for x,y in points)
        
        # get pen and brush
        pen = self._get_pen_attrs()
        brush = self._get_brush_attrs()
        
        # make command
        command = self._indent + '<polyline points="%s" %s %s fill-rule="evenodd" />' % (" ".join(points), pen, brush)
        
        # add command
        self._commands.append(command)
    
    
    def draw_path(self, path):
        """
        Draws the path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # get svg
        svg = path.svg(self._indent+_INDENT)
        
        # get pen and brush
        pen = self._get_pen_attrs()
        brush = self._get_brush_attrs()
        fill = SVG_FILL_RULE[path.fill_rule]
        
        # make command
        command = self._indent + '<path %s %s fill-rule="%s" d="%s" />' % (pen, brush, fill, svg)
        
        # add command
        self._commands.append(command)
    
    
    def draw_polygon(self, points):
        """
        Draws a closed polygon using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 3:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # format
        points = ("%s,%s" % (x,y) for x,y in points)
        
        # get pen and brush
        pen = self._get_pen_attrs()
        brush = self._get_brush_attrs()
        
        # make command
        command = self._indent + '<polygon points="%s" %s %s fill-rule="evenodd" />' % (" ".join(points), pen, brush)
        
        # add command
        self._commands.append(command)
    
    
    def draw_rect(self, x, y, width, height, radius=0):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corners starting from top-left.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # get radius
        if isinstance(radius, (int, float)) and radius != 0:
            radius = (radius, radius, radius, radius)
        
        # draw as path if different corners
        if radius and not all(r == radius[0] for r in radius):
            
            path = Path()
            path.rect(x, y, width, height, radius)
            self.draw_path(path)
            
            return
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # get pen and brush
        pen = self._get_pen_attrs()
        brush = self._get_brush_attrs()
        
        # no round corners
        if not radius:
            command = self._indent + '<rect x="%s" y="%s" width="%s" height="%s" %s %s />' % (x, y, width, height, pen, brush)
        
        # same radius for all corners
        else:
            radius = self._scale * radius[0]
            command = self._indent + '<rect x="%s" y="%s" width="%s" height="%s" rx="%s" ry="%s" %s %s />' % (x, y, width, height, radius, radius, pen, brush)
        
        # add command
        self._commands.append(command)
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # get font attributes
        font = self._get_text_attrs()
        
        # get full size
        full_width, full_height = self.get_text_size(text)
        descent = self._font_descent / self._scale
        
        # get rotation origin
        trans_x = self._scale * (x + self._offset[0])
        trans_y = self._scale * (y + self._offset[1])
        angle = numpy.rad2deg(angle)
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # draw lines
        for i, line in enumerate(lines):
            
            # get line size
            line_width, line_height = self.get_line_size(line)
            line_width /= self._scale
            line_height /= self._scale
            
            # init offset
            x_offset = 0
            y_offset = line_height - descent
            
            # adjust alignment
            if self.text_align == TEXT_ALIGN_CENTER:
                x_offset -= 0.5*line_width
            
            elif self.text_align == TEXT_ALIGN_RIGHT:
                x_offset -= line_width
            
            # adjust baseline
            if self.text_base == TEXT_BASE_MIDDLE:
                y_offset -= 0.5*full_height
            
            elif self.text_base == TEXT_BASE_BOTTOM:
                y_offset -= full_height
            
            # add line offset
            y_offset += i * line_height * (1 + self.text_spacing)
            
            # apply scaling and offset
            text_x = self._scale * (x + x_offset + self._offset[0])
            text_y = self._scale * (y + y_offset + self._offset[1])
            
            # escape line
            line = html.escape(line)
            
            # make command
            transform = 'transform="rotate(%s, %s, %s)"' % (angle, trans_x, trans_y) if angle else ""
            command = self._indent + '<text x="%s" y="%s" %s %s>%s</text>' % (text_x, text_y, font, transform, line)
            
            # add command
            self._commands.append(command)
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # make clip path
        name = "clip_%03d" % len(self._clips)
        self._clips[name] = '<clipPath id="%s"><path d="%s" /></clipPath>' % (name, path.svg())
        
        # make command
        command = self._indent + '<g clip-path="url(#%s)">' % name
        
        # add command
        self._commands.append(command)
        
        # increase indentation
        self._indent += _INDENT
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """Removes last clipping path while keeping previous if any."""
        
        # decrease indentation
        self._indent = self._indent[:-len(_INDENT)]
        
        # add command
        self._commands.append(self._indent + '</g>')
    
    
    def group(self, id_tag=None, class_tag=None):
        """
        Opens new drawing group.
        
        Args:
            id_tag: str
                Unique id of the group.
            
            class_tag:
                Class of the group.
        
        Returns:
            pero.GroupState
                Grouping state context manager.
        """
        
        # make command
        id_tag = ' id="%s"' % id_tag if id_tag else ""
        class_tag = ' class="%s"' % class_tag if class_tag else ""
        command = self._indent + '<g%s%s>' % (id_tag, class_tag)
        
        # add command
        self._commands.append(command)
        
        # increase indentation
        self._indent += _INDENT
        
        # return state
        return GroupState(self)
    
    
    def ungroup(self):
        """Closes the last drawing group."""
        
        # decrease indentation
        self._indent = self._indent[:-len(_INDENT)]
        
        # add command
        self._commands.append(self._indent + '</g>')
    
    
    def _get_pen_attrs(self):
        """Gets current pen attributes."""
        
        attrs = ('%s="%s"' % attr for attr in self._pen_attrs.items() if attr[1] is not None)
        return " ".join(attrs)
    
    
    def _get_brush_attrs(self):
        """Gets current brush attributes."""
        
        attrs = ('%s="%s"' % attr for attr in self._brush_attrs.items() if attr[1] is not None)
        return " ".join(attrs)
    
    
    def _get_text_attrs(self):
        """Gets current text attributes."""
        
        attrs = ('%s="%s"' % attr for attr in self._font_attrs.items() if attr[1] is not None)
        return " ".join(attrs)
    
    
    def _get_text_bgr_filter(self):
        """Gets the text background filter."""
        
        # get color
        color = ColorProperties.get_color(self, "text_bgr_")
        if not color or color.alpha == 0:
            return None
        
        # make name
        name = "text_bgr_%s" % color.hex[1:]
        
        # make filter if not present
        if name not in self._filters:
            
            flood = 'flood-color="%s"' % color.hex[:-2]
            if color.alpha != 255:
                flood += ' flood-opacity="%s"' % (color.alpha/255.)
            
            self._filters[name] = '<filter id="%s" x="0" y="0" width="1" height="1"><feFlood %s /><feComposite in="SourceGraphic" /></filter>' % (name, flood)
        
        return name
    
    
    def _update_pen(self, evt=None):
        """Updates pen with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('line_color', 'line_alpha'):
            color = ColorProperties.get_color(self, "line_")
            if color is not UNDEF:
                self._pen_attrs['stroke'] = color.hex[:-2]
                self._pen_attrs['stroke-opacity'] = color.alpha/255. if color.alpha != 255 else None
        
        # update width
        if prop_name is None or prop_name in ('line_width', 'line_scale'):
            line_width = self.line_width
            if line_width is not UNDEF:
                self._pen_attrs['stroke-width'] = line_width * self.line_scale
        
        # update cap
        if prop_name is None or prop_name == 'line_cap':
            line_cap = self.line_cap
            if line_cap is not UNDEF:
                self._pen_attrs['stroke-linecap'] = SVG_LINE_CAP[line_cap]
        
        # update join
        if prop_name is None or prop_name == 'line_join':
            line_join = self.line_join
            if line_join is not UNDEF:
                self._pen_attrs['stroke-linejoin'] = SVG_LINE_JOIN[line_join]
        
        # update style/dash
        if prop_name is None or prop_name in ('line_dash', 'line_style', 'line_width', 'line_scale'):
            line_style = self.line_style
            line_dash = self.line_dash if self.line_dash else []
            line_width = self._pen_attrs['stroke-width']
            
            if line_style == LINE_STYLE_SOLID:
                line_dash = []
            elif line_style not in (LINE_STYLE_CUSTOM, UNDEF):
                line_dash = SVG_LINE_STYLE[line_style]
            
            line_dash = ",".join(str(x*line_width) for x in line_dash) if line_dash else None
            
            self._pen_attrs['stroke-dasharray'] = line_dash
    
    
    def _update_brush(self, evt=None):
        """Updates brush with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('fill_color', 'fill_alpha', 'fill_style'):
            color = ColorProperties.get_color(self, "fill_")
            
            if self.fill_style == FILL_STYLE_TRANS:
                self._brush_attrs['fill'] = None
                self._brush_attrs['fill-opacity'] = 0
            
            elif color is not UNDEF:
                self._brush_attrs['fill'] = color.hex[:-2]
                self._brush_attrs['fill-opacity'] = color.alpha/255. if color.alpha != 255 else None
    
    
    def _update_text(self, evt=None):
        """Updates text with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # init flag
        font_changed = False
        
        # update font name/family
        if prop_name is None or prop_name in ('font_name', 'font_family'):
            font_name = self.font_name
            font_family = self.font_family
            
            if font_name:
                self._font_attrs['font-family'] = font_name
            elif font_family is None:
                self._font_attrs['font-family'] = None
            elif font_family is not UNDEF:
                self._font_attrs['font-family'] = SVG_FONT_FAMILY[font_family]
            
            font_changed = True
        
        # update font size
        if prop_name is None or prop_name in ('font_size', 'font_scale'):
            font_size = self.font_size
            
            if font_size is None:
                self._font_attrs['font-size'] = 11 * self.font_scale
            elif font_size is not UNDEF:
                self._font_attrs['font-size'] = font_size * self.font_scale
            
            font_changed = True
        
        # update font style
        if prop_name is None or prop_name == 'font_style':
            font_style = self.font_style
            
            if font_style is None:
                self._font_attrs['font-style'] = None
            elif font_style == FONT_STYLE_NORMAL:
                self._font_attrs['font-style'] = None
            elif font_style is not UNDEF:
                self._font_attrs['font-style'] = SVG_FONT_STYLE[font_style]
            
            font_changed = True
        
        # update font weight
        if prop_name is None or prop_name == 'font_weight':
            font_weight = self.font_weight
            
            if font_weight is None:
                self._font_attrs['font-weight'] = None
            elif font_weight == FONT_WEIGHT_NORMAL:
                self._font_attrs['font-weight'] = None
            elif font_weight is not UNDEF:
                self._font_attrs['font-weight'] = SVG_FONT_WEIGHT[font_weight]
            
            font_changed = True
        
        # update foreground color
        if prop_name is None or prop_name in ('text_color', 'text_alpha'):
            color = ColorProperties.get_color(self, "text_")
            if color is not UNDEF:
                self._font_attrs['fill'] = color.hex[:-2]
                self._font_attrs['fill-opacity'] = color.alpha/255. if color.alpha != 255 else None
        
        # update background color
        if prop_name is None or prop_name in ('text_bgr_color', 'text_bgr_alpha'):
            bgr_filter = self._get_text_bgr_filter()
            
            if bgr_filter:
                self._font_attrs['filter'] = 'url(#%s)' % bgr_filter
            else:
                self._font_attrs['filter'] = None
        
        # get final font family
        if font_changed:
            font = self.get_font()
            self._font_attrs['font-family'] = font.name
            self._font_descent = font.get_descent(int(0.5+self._font_attrs['font-size']))



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for seconds formatter."""
    
    
    def test_standalone(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.SecondsFormatter()
        
        # test days
        self.assertEqual(formatter.format(24*60*60), "1 d")
        self.assertEqual(formatter.format(24*60*60 + 30*60), "1 d")
        
        # test hours
        self.assertEqual(formatter.format(60*60), "1 h")
        self.assertEqual(formatter.format(24*60*60 - 1), "24 h")
        
        # test minutes
        self.assertEqual(formatter.format(60), "1 m")
        self.assertEqual(formatter.format(60*60 - 1), "60 m")
        
        # test seconds
        self.assertEqual(formatter.format(1), "1 s")
        self.assertEqual(formatter.format(60-1), "59 s")
        
        # test milliseconds
        self.assertEqual(formatter.format(1e-3), "1 ms")
        self.assertEqual(formatter.format(1-0.1), "900 ms")
        
        # test microseconds
        self.assertEqual(formatter.format(1e-6), "1 us")
        self.assertEqual(formatter.format(1e-3-1e-4), "900 us")
    
    
    def test_places(self):
        """Tests whether places works correctly."""
        
        formatter = pero.SecondsFormatter(places=0)
        
        self.assertEqual(formatter.format(60*60 + 30*60), "2 h")
        self.assertEqual(formatter.format(70.5), "1 m")
        self.assertEqual(formatter.format(17.567), "18 s")
        self.assertEqual(formatter.format(.1), "100 ms")
        
        formatter = pero.SecondsFormatter(places=2)
        
        self.assertEqual(formatter.format(60*60 + 30*60), "1.50 h")
        self.assertEqual(formatter.format(70.5), "1.18 m")
        self.assertEqual(formatter.format(17.567), "17.57 s")
        self.assertEqual(formatter.format(.1), "100.00 ms")
    
    
    def test_precision(self):
        """Tests whether precision works correctly."""
        
        formatter = pero.SecondsFormatter(precision=60*60)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "1 h")
        self.assertEqual(formatter.format(70.5), "1 m")
        self.assertEqual(formatter.format(17.567), "18 s")
        self.assertEqual(formatter.format(.1), "100 ms")
        
        formatter = pero.SecondsFormatter(precision=60)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "1.50 h")
        self.assertEqual(formatter.format(70.5), "1 m")
        self.assertEqual(formatter.format(17.567), "18 s")
        self.assertEqual(formatter.format(.1), "100 ms")
    
    
    def test_domain(self):
        """Tests whether domain works correctly."""
        
        formatter = pero.SecondsFormatter(domain=60*60)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "1 h")
        self.assertEqual(formatter.format(70.5), "0 h")
        self.assertEqual(formatter.format(17.567), "0 h")
        self.assertEqual(formatter.format(.1), "0 h")
        
        formatter = pero.SecondsFormatter(domain=60)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "90 m")
        self.assertEqual(formatter.format(70.5), "1 m")
        self.assertEqual(formatter.format(17.567), "0 m")
        self.assertEqual(formatter.format(.1), "0 m")
    
    
    def test_digits(self):
        """Tests whether digits works correctly."""
        
        # check domain and precision
        formatter = pero.SecondsFormatter(domain=60*60, precision=60)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "1.50 h")
        self.assertEqual(formatter.format(70.5), "0.02 h")
        self.assertEqual(formatter.format(17.567), "0.00 h")
        self.assertEqual(formatter.format(.1), "0.00 h")
        
        formatter = pero.SecondsFormatter(domain=60*4, precision=30)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "90.0 m")
        self.assertEqual(formatter.format(70.5), "1.2 m")
        self.assertEqual(formatter.format(17.567), "0.3 m")
        self.assertEqual(formatter.format(.1), "0.0 m")
        
        # check domain, precision and places
        formatter = pero.SecondsFormatter(domain=60*60, precision=60, places=1)
        
        self.assertEqual(formatter.format(60*60 + 30*60 - 1), "1.5 h")
        self.assertEqual(formatter.format(70.5), "0.0 h")
        self.assertEqual(formatter.format(17.567), "0.0 h")
        self.assertEqual(formatter.format(.1), "0.0 h")
    
    
    def test_suffix(self):
        """Tests whether suffix works correctly."""
        
        # standalone formatter
        formatter = pero.SecondsFormatter(hide_suffix=True)
        self.assertEqual(formatter.suffix(), "")
        
        self.assertEqual(formatter.format(60*60), "1")
        self.assertEqual(formatter.suffix(), " (h)")
        
        self.assertEqual(formatter.format(60), "1")
        self.assertEqual(formatter.suffix(), " (m)")
        
        self.assertEqual(formatter.format(59), "59")
        self.assertEqual(formatter.suffix(), " (s)")
        
        # formatter with domain
        formatter = pero.SecondsFormatter(hide_suffix=True, domain=60)
        self.assertEqual(formatter.suffix(), " (m)")
        
        self.assertEqual(formatter.format(60*60), "60")
        self.assertEqual(formatter.suffix(), " (m)")
        
        self.assertEqual(formatter.format(60), "1")
        self.assertEqual(formatter.suffix(), " (m)")
        
        self.assertEqual(formatter.format(59), "1")
        self.assertEqual(formatter.suffix(), " (m)")
        
        self.assertEqual(formatter.format(25), "0")
        self.assertEqual(formatter.suffix(), " (m)")
    
    
    def test_convert(self):
        """Tests whether convert works correctly."""
        
        formatter = pero.SecondsFormatter(domain=60)
        
        self.assertEqual(formatter.convert(60*60), 60)
        self.assertEqual(formatter.convert(60), 1)
        self.assertEqual(formatter.convert(30), 0.5)
        self.assertEqual(formatter.convert(15), 0.25)
        self.assertEqual(formatter.convert(0.6), 0.01)
    
    
    def test_invert(self):
        """Tests whether invert works correctly."""
        
        formatter = pero.SecondsFormatter(domain=60)
        
        self.assertEqual(formatter.invert(60), 60*60)
        self.assertEqual(formatter.invert(1), 60)
        self.assertEqual(formatter.invert(0.5), 30)
        self.assertEqual(formatter.invert(0.25), 15)
        self.assertEqual(formatter.invert(0.01), 0.6)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for Palette class."""
    
    
    def test_constructor(self):
        """Tests whether constructor works correctly."""
        
        color0 = pero.Color.Red
        color1 = pero.Color(100, 150, 200)
        color2 = "#aaa"
        
        palette = pero.Palette((color0, color1, color2))
    
    
    def test_getitem(self):
        """Tests whether colors are accessed correctly."""
        
        color0 = pero.Color.Red
        color1 = pero.Color.Green
        color2 = pero.Color.Blue
        
        palette = pero.Palette((color0, color1, color2))
        
        self.assertTrue(palette[0] is color0)
        self.assertTrue(palette[1] is color1)
        self.assertTrue(palette[2] is color2)
    
    
    def test_reversed(self):
        """Tests whether reversed palette is correctly created from parent."""
        
        color0 = pero.Color.Red
        color1 = pero.Color.Green
        color2 = pero.Color.Blue
        color3 = pero.Color.Cyan
        color4 = pero.Color.Magenta
        color5 = pero.Color.Yellow
        
        colors = (color0, color1, color2, color3, color4, color5)
        palette = pero.Palette(colors)
        
        rev = palette.reversed("Reversed")
        
        self.assertEqual(rev.name, "Reversed")
        self.assertEqual(len(rev), len(palette))
        
        self.assertTrue(rev[0] is palette[5])
        self.assertTrue(rev[1] is palette[4])
        self.assertTrue(rev[2] is palette[3])
        self.assertTrue(rev[3] is palette[2])
        self.assertTrue(rev[4] is palette[1])
        self.assertTrue(rev[5] is palette[0])
    
    
    def test_from_name(self):
        """Tests whether named palettes can be accessed."""
        
        # test from name
        palette = pero.Palette.from_name('Spectral')
        palette = pero.Palette.from_name('spectral')
        
        # test from class
        palette = pero.Palette.Spectral
        palette = pero.Palette.spectral
        
        # test from lib
        palette = pero.PALETTES.Spectral
        palette = pero.PALETTES.spectral
        
        # test from module
        palette = pero.colors.Spectral
    
    
    def test_from_palette(self):
        """Tests whether new palette is correctly created from parent."""
        
        color0 = pero.Color.Red
        color1 = pero.Color.Green
        color2 = pero.Color.Blue
        color3 = pero.Color.Cyan
        color4 = pero.Color.Magenta
        color5 = pero.Color.Yellow
        color6 = pero.Color.Black
        
        colors = (color0, color1, color2, color3, color4, color5, color6)
        palette = pero.Palette(colors)
        
        picked = pero.Palette.from_palette(palette, 2)
        self.assertEqual(len(picked), 2)
        self.assertTrue(picked[0] is color0)
        self.assertTrue(picked[1] is color6)
        
        picked = pero.Palette.from_palette(palette, 3)
        self.assertEqual(len(picked), 3)
        self.assertTrue(picked[0] is color0)
        self.assertTrue(picked[1] is color3)
        self.assertTrue(picked[2] is color6)
        
        picked = pero.Palette.from_palette(palette, 4)
        self.assertEqual(len(picked), 4)
        self.assertTrue(picked[0] is color0)
        self.assertTrue(picked[1] is color2)
        self.assertTrue(picked[2] is color4)
        self.assertTrue(picked[3] is color6)
    
    
    def test_from_gradient(self):
        """Tests whether new palette is correctly created from gradient."""
        
        color0 = pero.Color.Black
        color1 = pero.Color.White
        
        colors = (color0, color1)
        gradient = pero.Gradient(colors)
        
        picked = pero.Palette.from_gradient(gradient, 2)
        self.assertEqual(len(picked), 2)
        self.assertEqual(picked[0], color0)
        self.assertEqual(picked[1], color1)
        
        picked = pero.Palette.from_gradient(gradient, 3)
        self.assertEqual(len(picked), 3)
        self.assertEqual(picked[0], color0)
        self.assertEqual(picked[1], (128, 128, 128))
        self.assertEqual(picked[2], color1)
        
        picked = pero.Palette.from_gradient(gradient, 4)
        self.assertEqual(len(picked), 4)
        self.assertEqual(picked[0], color0)
        self.assertEqual(picked[1], (85, 85, 85))
        self.assertEqual(picked[2], (170, 170, 170))
        self.assertEqual(picked[3], color1)
    
    
    def test_name_registering(self):
        """Tests whether new named palettes are registered."""
        
        # init named palette
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        model = pero.Palette(colors, "MyPalette")
        
        # test from name
        palette = pero.Palette.from_name('MyPalette')
        self.assertTrue(palette is model)
        
        palette = pero.Palette.from_name('mypalette')
        self.assertTrue(palette is model)
        
        # test from class
        palette = pero.Palette.MyPalette
        self.assertTrue(palette is model)
        
        palette = pero.Palette.mypalette
        self.assertTrue(palette is model)
        
        # test from lib
        palette = pero.PALETTES.MyPalette
        self.assertTrue(palette is model)
        
        palette = pero.PALETTES.mypalette
        self.assertTrue(palette is model)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . formatter import Formatter
from . funced import FuncFormatter
from . indexed import IndexFormatter
from . scalar import ScalarFormatter
from . spliter import SplitFormatter, EngFormatter, BytesFormatter, SecondsFormatter
from . stringed import EmptyFormatter, StrFormatter, PrintfFormatter
from . timed import TimeFormatter

# import utils
from . utils import *



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero
import numpy


class TestCase(unittest.TestCase):
    """Test case for linear interpolator."""
    
    
    def test_log10(self):
        """Tests whether interpolator works for log10 range."""
        
        interpol = pero.LogInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(10, 1, 100), 0.5)
        self.assertEqual(interpol.denormalize(0.5, 1, 100), 10)
        
        # test left
        self.assertAlmostEqual(interpol.normalize(0.1, 1, 100), -0.5, 10)
        self.assertAlmostEqual(interpol.denormalize(-0.5, 1, 100), 0.1, 10)
        
        # test right
        self.assertAlmostEqual(interpol.normalize(1000, 1, 100), 1.5, 10)
        self.assertAlmostEqual(interpol.denormalize(1.5, 1, 100), 1000, 10)
    
    
    def test_log2(self):
        """Tests whether interpolator works for log2 range."""
        
        interpol = pero.LogInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(2, 1, 4), 0.5)
        self.assertEqual(interpol.denormalize(0.5, 1, 4), 2)
        
        # test left
        self.assertAlmostEqual(interpol.normalize(0.5, 1, 4), -0.5, 10)
        self.assertAlmostEqual(interpol.denormalize(-0.5, 1, 4), 0.5, 10)
        
        # test right
        self.assertAlmostEqual(interpol.normalize(8, 1, 4), 1.5, 10)
        self.assertAlmostEqual(interpol.denormalize(1.5, 1, 4), 8, 10)
    
    
    def test_arrays(self):
        """Tests whether interpolator works correctly with arrays."""
        
        interpol = pero.LogInterpol()
        
        data = [0.1, 10, 100, 1000]
        model = [-0.5, 0.5, 1, 1.5]
        
        numpy.testing.assert_almost_equal(list(interpol.normalize(numpy.array(data), 1, 100)), model, 10)
        numpy.testing.assert_almost_equal(list(interpol.denormalize(numpy.array(model), 1, 100)), data, 10)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import Layout
from . control import Control


class Sizer(Control):
    """
    Sizer encapsulates the pero.Layout and relays all interactive events and
    drawing mechanism down to individual cells. It allows interactive controls
    to be drawn in specific layout.
    
    Properties:
        
        graphics: pero.Layout
            Layout drawing tool.
    """
    
    graphics = Property(UNDEF, types=Layout, dynamic=False, nullable=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Sizer."""
        
        # init base
        super().__init__(**overrides)
        
        # init buffs
        self._focused = None
        
        # init layout
        if not self.graphics:
            self.graphics = Layout()
        
        # lock layout
        self.lock_property('graphics')
        
        # bind events
        self.bind(EVT_SIZE, self._on_sizer_event)
        self.bind(EVT_KEY_DOWN, self._on_sizer_event)
        self.bind(EVT_KEY_UP, self._on_sizer_event)
        self.bind(EVT_MOUSE_ENTER, self._on_sizer_event)
        self.bind(EVT_MOUSE_LEAVE, self._on_sizer_event)
        self.bind(EVT_MOUSE_MOTION, self._on_sizer_event)
        self.bind(EVT_MOUSE_SCROLL, self._on_sizer_event)
        self.bind(EVT_LEFT_DOWN, self._on_sizer_event)
        self.bind(EVT_LEFT_UP, self._on_sizer_event)
        self.bind(EVT_LEFT_DCLICK, self._on_sizer_event)
        self.bind(EVT_RIGHT_DOWN, self._on_sizer_event)
        self.bind(EVT_RIGHT_UP, self._on_sizer_event)
        self.bind(EVT_TOUCH_START, self._on_sizer_event)
        self.bind(EVT_TOUCH_END, self._on_sizer_event)
        self.bind(EVT_TOUCH_MOVE, self._on_sizer_event)
        self.bind(EVT_TOUCH_CANCEL, self._on_sizer_event)
        self.bind(EVT_TOUCH_DTAP, self._on_sizer_event)
    
    
    def get_cell(self, row, col):
        """
        Gets the cell at specified layout grid position.
        
        Args:
            row: int
                Row index of requested cell.
            
            col: int
                Column index of requested cell.
        
        Returns:
            pero.Cell or None
                Corresponding cell or None.
        """
        
        return self.graphics.get_cell(row, col)
    
    
    def get_cell_below(self, x, y):
        """
        Gets the cell for which given coordinates fall into its bounding box.
        
        Args:
            x: int or float
                X-coordinate in logical units.
            
            y: int or float
                Y-coordinate in logical units.
        
        Returns:
            pero.Cell or None
                Corresponding cell or None.
        """
        
        return self.graphics.get_cell_below(x, y)
    
    
    def add(self, graphics, row, col, **overrides):
        """
        Adds graphics to specified layout cell. Additional rows and columns are
        added if necessary with relative size of 1. See pero.Layout for more.
        
        Args:
            graphics: pero.Graphics
                Graphics to be added.
            
            row: int
                Index of the row into which the graphics should be added.
            
            col: int
                Index of the column into which the graphics should be added.
            
            overrides: (str:?)
                Additional cell properties.
        """
        
        # set parent for controls
        if isinstance(graphics, Control):
            graphics.set_parent(self)
        
        # add to layout
        self.graphics.add(graphics, row, col, **overrides)
    
    
    def add_row(self, height=1, relative=True):
        """
        Adds additional row at the end of layout with specified absolute or
        relative height.
        
        Args:
            height: int or float
                Absolute or relative height of the row.
            
            relative:
                If set to True, specified height is considered as relative
                portion of total available space after filling all fixed rows.
        """
        
        # add to layout
        self.graphics.add_row(height, relative)
    
    
    def add_col(self, width=1, relative=True):
        """
        Adds additional column at the end of layout with specified absolute or
        relative width.
        
        Args:
            width: int or float
                Absolute or relative width of the column.
            
            relative:
                If set to True, specified width is considered as relative
                portion of total available space after filling all fixed
                columns.
        """
        
        # add to layout
        self.graphics.add_col(width, relative)
    
    
    def draw_overlay(self, func=None, view=None, **kwargs):
        """Relays overlay drawing to parent."""
        
        # check parent
        if self._parent is None:
            return
        
        # get cell view shift
        view = None
        if self._focused:
            view = self._focused.content.rect
        
        # draw overlay by parent
        self._parent.draw_overlay(func, view, **kwargs)
    
    
    def _on_sizer_event(self, evt):
        """Relay events to focused control."""
        
        # process cursor event
        if hasattr(evt, 'x_pos') and hasattr(evt, 'y_pos'):
            
            # get item under cursor
            self._focused = self.graphics.get_cell_below(evt.x_pos, evt.y_pos)
            if self._focused is None:
                return
            
            # reset event position
            evt.x_pos, evt.y_pos = self._focused.to_content(evt.x_pos, evt.y_pos)
        
        # skip if no cell in focus
        if self._focused is None:
            return
        
        # reset event control
        if hasattr(evt, 'control'):
            evt.control = self._focused.graphics
        
        # raise event in focused item
        self._focused.graphics.fire(evt)
